// This is WXGLMAIN.CPP. Copyright (C) 2004 by Pawel Pilarczyk.
// It is a framework program for the wxWindows library and Open GL.
// This is free software. No warranty. Consult 'license.txt' for details.
// Started on June 30, 2004. Last revision: August 13, 2004.


// #define _WXGLMAIN_CPP_

//#include "options.h"
#include "wxglmain.h"

// avoid the re-definition of _puttchar warning in wx 2.5.2 (MSW/MinGW)
#if defined (__WXMSW__) && defined (__MINGW32__)
#  include <tchar.h>
#endif

// include the declarations of sin and cos
#include <math.h>

// include the declarations for setting locale (LC_NUMERIC in particular)
#include <locale.h>

// include the wxWindows header files
#include <wx/wx.h>
#include <wx/image.h>
#include <wx/imagpng.h>
#include <wx/glcanvas.h>
#include <wx/colordlg.h>

const char *proginfo =
	"This program displays 3D objects with the use of OpenGL. "
	"Use left mouse button or arrows to rotate the figure, "
	"right mouse button or ADSW to move to the sides, "
	"middle mouse button or +/- to get closer/farther, "
	"or hit Home or 0 to reset the viewing position.";

const char *defaultStatusText =
	"This program uses wxWindows and OpenGL.";

const char *aboutinfo =
	"This is a wxWindows program that uses OpenGL "
	"in a framework programmed in 2004 by Pawel Pilarczyk. "
	"This is free software. No warranty. "
	"See the GNU General Public License for details.";
	
// the rotation and translation matrices
float rotation [16];
float translation [3] = {0.0f, 0.0f, 0.0f};

// the background color of the scene
float bgcolor [4] = {0.0f, 0.0f, 0.0f, 0.0f};

// the magnification exponent
double magnification = 0.0;

// should the scene be plotted in the wire-frame mode?
bool wireframe = false;

// should there be a fog in the scene?
bool fog = true;

// should the perspective be used in the projection?
bool perspective = true;


// --------------------------------------------------
// ---------------- wxGL Main Canvas ----------------
// --------------------------------------------------
// The OpenGL canvas contains the OpenGL graphics
// and is contained in the main frame.

class KrakCanvas: public wxGLCanvas
{
public:
	KrakCanvas (wxWindow *parent, wxWindowID id = wxID_ANY,
		const wxPoint &pos = wxDefaultPosition,
		const wxSize &size = wxDefaultSize, long style = 0,
		const wxString &name = "Krak Canvas",
		int *gl_attrib = NULL);

	~KrakCanvas ();

	void OnPaint (wxPaintEvent &event);
	void OnSize (wxSizeEvent &event);
	
	void OnEraseBackground (wxEraseEvent &event);

	void OnMouseEvents (wxMouseEvent &event);
	void OnChar (wxKeyEvent &event);

	void Rotate (int x, int y);
	void Rescale (int diff, int x, int y);
	
	DECLARE_EVENT_TABLE()
}; /* KrakCanvas */

KrakCanvas *krakCanvas = (KrakCanvas *) NULL;

// --------------------------------------------------

BEGIN_EVENT_TABLE (KrakCanvas, wxGLCanvas)
	EVT_MOUSE_EVENTS (KrakCanvas::OnMouseEvents)
	EVT_PAINT (KrakCanvas::OnPaint)
	EVT_SIZE (KrakCanvas::OnSize)
#ifndef __WXGTK__
	EVT_CHAR (KrakCanvas::OnChar)
#endif
	EVT_ERASE_BACKGROUND (KrakCanvas::OnEraseBackground)
END_EVENT_TABLE ()


// --------------------------------------------------
// ---------------- wxGL Main Frame -----------------
// --------------------------------------------------
// This class is the main frame in the program.
// It contains the OpenGL canvas, and a panel which receives
// the keyboard events (the latter is necessary only with GTK).

class KrakFrame: public wxFrame
{
public:
	bool showScaling;
	bool showKeys;
	bool showMouse;
	bool showRefresh;
	
	// frame constructor and destructor
	KrakFrame (wxWindow *parent, const wxString &title,
		const wxPoint &pos, const wxSize &size);
	~KrakFrame ();

	// event handling
	void OnQuit (wxCommandEvent &event);
	void Save (const char *files, wxBitmapType type);
	void OnSaveBMP (wxCommandEvent &event);
	void OnSavePNG (wxCommandEvent &event);
	void OnShowWireFrame (wxCommandEvent &event);
	void OnShowFog (wxCommandEvent &event);
	void OnPerspective (wxCommandEvent &event);
	void OnBgColor (wxCommandEvent& event);
	void OnShowScaling (wxCommandEvent &event);
	void OnShowKeys (wxCommandEvent &event);
	void OnShowMouse (wxCommandEvent &event);
	void OnShowRefresh (wxCommandEvent &event);
	void OnAbout (wxCommandEvent &event);
	void OnHelpGeneral (wxCommandEvent &event);
	void OnHelpSpecific (wxCommandEvent &event);
	void OnSize (wxSizeEvent &event);
	void OnClose (wxCloseEvent& event);

	void OnChar (wxKeyEvent &event)
	{
		if (krakCanvas)
			krakCanvas -> OnChar (event);
		return;
	}
	
	DECLARE_EVENT_TABLE ()

}; /* class KrakFrame */

KrakFrame *krakFrame = (KrakFrame *) NULL;

// --------------------------------------------------

enum
{
	ID_Quit = 1,
	ID_SaveBMP,
	ID_SavePNG,
	ID_About,
	ID_HelpGeneral,
	ID_HelpSpecific,
	ID_ShowWireFrame,
	ID_ShowFog,
	ID_Perspective,
	ID_BgColor,
	ID_ShowScaling,
	ID_ShowKeys,
	ID_ShowMouse,
	ID_ShowRefresh
};

BEGIN_EVENT_TABLE (KrakFrame, wxFrame)
	EVT_MENU (ID_Quit, KrakFrame::OnQuit)
	EVT_MENU (ID_SaveBMP, KrakFrame::OnSaveBMP)
	EVT_MENU (ID_SavePNG, KrakFrame::OnSavePNG)
	EVT_MENU (ID_HelpGeneral, KrakFrame::OnHelpGeneral)
	EVT_MENU (ID_HelpSpecific, KrakFrame::OnHelpSpecific)
	EVT_MENU (ID_About, KrakFrame::OnAbout)
	EVT_MENU (ID_ShowWireFrame, KrakFrame::OnShowWireFrame)
	EVT_MENU (ID_ShowFog, KrakFrame::OnShowFog)
	EVT_MENU (ID_Perspective, KrakFrame::OnPerspective)
	EVT_MENU (ID_BgColor, KrakFrame::OnBgColor)
	EVT_MENU (ID_ShowScaling, KrakFrame::OnShowScaling)
	EVT_MENU (ID_ShowKeys, KrakFrame::OnShowKeys)
	EVT_MENU (ID_ShowMouse, KrakFrame::OnShowMouse)
	EVT_MENU (ID_ShowRefresh, KrakFrame::OnShowRefresh)
	EVT_SIZE (KrakFrame::OnSize)
#ifndef __WXGTK__
	EVT_CHAR (KrakFrame::OnChar)
#endif
	EVT_CLOSE (KrakFrame::OnClose)
END_EVENT_TABLE ()


#ifdef __WXGTK__
// --------------------------------------------------
// ------------------- Krak Panel -------------------
// --------------------------------------------------
// This is a panel that receives keyboard events in GTK.

class KrakPanel: public wxPanel
{
public:
	KrakPanel (wxWindow *parent): wxPanel (parent, -1,
		wxPoint (0, 0), wxSize (0, 0)) {return;}

	void OnChar (wxKeyEvent &event)
	{
		if (krakCanvas)
			krakCanvas -> OnChar (event);
		return;
	}
	
	DECLARE_EVENT_TABLE ()

}; /* class KrakPanel */

KrakPanel *krakPanel = (KrakPanel *) NULL;

// --------------------------------------------------

BEGIN_EVENT_TABLE (KrakPanel, wxPanel)
	EVT_CHAR (KrakPanel::OnChar)
END_EVENT_TABLE ()

#endif // __WXGTK__


// --------------------------------------------------
// ---------------- wxGL Application ----------------
// --------------------------------------------------
// This is the main application.

class KrakApp: public wxApp
{
virtual bool OnInit ();
virtual int OnExit ();

}; /* class KrakApp */

// --------------------------------------------------

IMPLEMENT_APP (KrakApp)


// --------------------------------------------------
// ---------------- wxGL Main Frame -----------------
// --------------------------------------------------

KrakFrame::KrakFrame (wxWindow *parent, const wxString& title,
	const wxPoint& pos, const wxSize& size):
	wxFrame ((wxWindow *) NULL, -1, title, pos, size)
{
	// switches for showing additional info
	showScaling = false;
	showKeys = false;
	showMouse = false;
	showRefresh = false;

	// StatusBar
	CreateStatusBar ();
	SetStatusText (defaultStatusText);

	// the File menu
	wxMenu *menuFile = new wxMenu;
	menuFile -> Append (ID_SaveBMP, "&Save BMP",
		"Save the image to a BMP file");
	menuFile -> Append (ID_SavePNG, "&Save PNG",
		"Save the image to a PNG file");
	menuFile -> AppendSeparator ();
	menuFile -> Append (ID_Quit, "E&xit", "Exit the program");

	// the Options menu
	wxMenu *menuOptions = new wxMenu;
	menuOptions -> AppendCheckItem (ID_ShowWireFrame, "&Wireframe",
		"Show the scene as wireframe");
	menuOptions -> Check (ID_ShowWireFrame, wireframe);
	menuOptions -> AppendCheckItem (ID_ShowFog, "&Fog",
		"Add the fog effect to the scene");
	menuOptions -> Check (ID_ShowFog, fog);
	menuOptions -> AppendCheckItem (ID_Perspective, "&Perspective",
		"Use perspective in the scene");
	menuOptions -> Check (ID_Perspective, perspective);
	menuOptions -> AppendSeparator ();
	menuOptions -> Append (ID_BgColor, "&Background",
		"Change background color");

	// the Info menu
	wxMenu *menuInfo = new wxMenu;
	menuInfo -> AppendCheckItem (ID_ShowScaling, "&Magnification",
		"Show the scaling exponent");
	menuInfo -> AppendCheckItem (ID_ShowKeys, "&Keys",
		"Show the keys pressed");
	menuInfo -> AppendCheckItem (ID_ShowMouse, "&Mouse",
		"Show the mouse position and status");
	menuInfo -> AppendCheckItem (ID_ShowRefresh, "&Refresh",
		"Show the screen refresh counter");

	// the Help menu
	wxMenu *menuHelp = new wxMenu;
	menuHelp -> Append (ID_HelpGeneral, "&General...",
		"Show general help information");
	if (helpinfo && *helpinfo)
		menuHelp -> Append (ID_HelpSpecific, "&Specific...",
		"Show specific help information");
	menuHelp -> AppendSeparator ();
	menuHelp -> Append (ID_About, "&About...", "About the program");

	// the menu bar
	wxMenuBar *menuBar = new wxMenuBar;
	menuBar -> Append (menuFile, "&File");
	menuBar -> Append (menuOptions, "&Options");
	menuBar -> Append (menuInfo, "&Info");
	menuBar -> Append (menuHelp, "&Help");

	SetMenuBar (menuBar);

	// recalculate the new size
//	SendSizeEvent ();

	// show the frame
//	Show (TRUE);
	
	// other initializations

	// paint the initial screen
//	Refresh ();
	
	return;
} /* KrakFrame::KrakFrame */

KrakFrame::~KrakFrame ()
{
	return;
} /* KrakFrame::~KrakFrame */

void KrakFrame::OnQuit (wxCommandEvent& WXUNUSED (event))
{
	Close (TRUE);
	return;
} /* KrakFrame::OnQuit */

void KrakFrame::Save (const char *files, wxBitmapType type)
{
	// show a dialog box and ask for a file name to save the image to
	wxFileDialog *dlg = new wxFileDialog (this, "Save bitmap", "", "",
		files, wxSAVE | wxOVERWRITE_PROMPT);
	if (dlg -> ShowModal () != wxID_OK)
	{
		dlg -> Destroy ();
		return;
	}

	// capture the OpenGL graphics to a bitmap buffer
	GLint viewport [4];	// viewport: x, y, width, height
	glGetIntegerv (GL_VIEWPORT, viewport);
	int width = viewport [2];
	int height = viewport [3];
	int RGBwidth = width * 3;
	int BMPwidth = (RGBwidth + 3) & ~3;
	int BMPsize = BMPwidth * height;
	// BMPbuf must be allocated with "malloc": it is free'd in wxImage
	char *BMPbuf = (char *) malloc (BMPsize);
	if (!BMPbuf)
	{
		wxMessageBox ("Can't allocate memory for the bitmap buffer.",
			"Not saved", wxOK, this);
		dlg -> Destroy ();
		return;
	}
	glPixelStorei (GL_PACK_ALIGNMENT, 4);
	glPixelStorei (GL_PACK_ROW_LENGTH, 0);
	glPixelStorei (GL_PACK_SKIP_ROWS, 0);
	glPixelStorei (GL_PACK_SKIP_PIXELS, 0);
	glReadPixels (0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, BMPbuf);

	// remove gaps from the buffer if there are any
	if (BMPwidth != RGBwidth)
	{
		int height = viewport [3];
		char *src = BMPbuf + BMPwidth, *dst = BMPbuf + RGBwidth;
		if (!height)
			++ height;
		while (-- height)
		{
			for (int i = 0; i < RGBwidth; i ++)
				*(dst ++) = *(src ++);
			src += BMPwidth - RGBwidth;
		}
	}

	// turn the picture in the buffer upside down
	char *row1 = BMPbuf;
	char *row2 = BMPbuf + (height - 1) * RGBwidth;
	while (row1 < row2)
	{
		for (int i = 0; i < RGBwidth; i ++)
		{
			char tmp = *row1;
			*(row1 ++) = *row2;
			*(row2 ++) = tmp;
		}
		row2 -= RGBwidth;
		row2 -= RGBwidth;
	}

	// transform the buffer into a system-independent RGB image
	// note: BMPbuf will be free'd in the wxImage destructor
	wxImage BMPimg (width, height, (unsigned char *) BMPbuf, false);
	if (!BMPimg. Ok ())
	{
		wxMessageBox ("Cannot create a wxImage from the raw data.",
			"Not saved", wxOK, this);
		dlg -> Destroy ();
		return;
	}

	// save the bitmap image to a file
	bool result = BMPimg. SaveFile (dlg -> GetPath (), type);
	if (!result)
		wxMessageBox ("Unable to save - an error occurred.",
			"Not saved", wxOK, this);

	dlg -> Destroy ();
	return;
} /* KrakFrame::Save */

void KrakFrame::OnSaveBMP (wxCommandEvent& WXUNUSED (event))
{
	Save ("Bitmap files (*.bmp)|*.bmp|All files (*.*)|*.*",
		wxBITMAP_TYPE_BMP);
	return;
} /* KrakFrame::OnSaveBMP */

void KrakFrame::OnSavePNG (wxCommandEvent& WXUNUSED (event))
{
	static bool added = false;
	if (!added)
	{
		#if wxUSE_LIBPNG
		wxImage::AddHandler (new wxPNGHandler);
		#endif
		added = true;
	}
	Save ("PNG bitmap files (*.png)|*.png|All files (*.*)|*.*",
		wxBITMAP_TYPE_PNG);
	return;
} /* KrakFrame::OnSavePNG */

void KrakFrame::OnShowWireFrame (wxCommandEvent& event)
{
	if (event. IsChecked ())
		wireframe = true;
	else
		wireframe = false;

	krakCanvas -> Refresh ();
	return;
} /* KrakFrame::OnShowWireFrame */

void KrakFrame::OnShowFog (wxCommandEvent& event)
{
	if (event. IsChecked ())
		fog = true;
	else
		fog = false;

	krakCanvas -> Refresh ();
	return;
} /* KrakFrame::OnShowFog */

void KrakFrame::OnPerspective (wxCommandEvent& event)
{
	if (event. IsChecked ())
		perspective = true;
	else
		perspective = false;

	krakCanvas -> Rescale (0, 0, 0);
	krakCanvas -> Refresh ();
	return;
} /* KrakFrame::OnPerspective */

inline unsigned char cnv (float x)
{
	return (unsigned char) (x * 255);
} /* cnv */

inline float cnv (unsigned char x)
{
	return (float) (x / 255.0);
} /* cnv */

void KrakFrame::OnBgColor (wxCommandEvent& WXUNUSED (event))
{
	wxColor bgColor (cnv (bgcolor [0]), cnv (bgcolor [1]),
		cnv (bgcolor [2]));
	wxColourData cd;
	cd. SetColour (bgColor);
	wxColourDialog *dlg = new wxColourDialog (this, &cd);
	if (dlg -> ShowModal () == wxID_OK)
	{
		cd = dlg -> GetColourData ();
		bgColor = cd. GetColour ();
		bgcolor [0] = cnv (bgColor. Red ());
		bgcolor [1] = cnv (bgColor. Green ());
		bgcolor [2] = cnv (bgColor. Blue ());
		krakCanvas -> Refresh ();
	}
	dlg -> Destroy ();
	return;
} /* KrakFrame::OnBgColor */

inline void showScalingStatus (void)
{
	wxString msg;
	msg. Printf ("Magnification exponent = %.2f.", magnification);
	krakFrame -> SetStatusText (msg);
	return;
} /* showScalingStatus */

void KrakFrame::OnShowScaling (wxCommandEvent& event)
{
	if (event. IsChecked ())
	{
		showScaling = true;
		showScalingStatus ();
	}
	else
	{
		showScaling = false;
		SetStatusText (defaultStatusText);
	}
	return;
} /* KrakFrame::OnShowScaling */

void KrakFrame::OnShowKeys (wxCommandEvent& event)
{
	if (event. IsChecked ())
		showKeys = true;
	else
	{
		showKeys = false;
		SetStatusText (defaultStatusText);
	}
	return;
} /* KrakFrame::OnShowKeys */

void KrakFrame::OnShowMouse (wxCommandEvent& event)
{
	if (event. IsChecked ())
		showMouse = true;
	else
	{
		showMouse = false;
		SetStatusText (defaultStatusText);
	}
	return;
} /* KrakFrame::OnShowMouse */

void KrakFrame::OnShowRefresh (wxCommandEvent& event)
{
	if (event. IsChecked ())
		showRefresh = true;
	else
	{
		showRefresh = false;
		SetStatusText (defaultStatusText);
	}
	return;
} /* KrakFrame::OnShowRefresh */

void KrakFrame::OnAbout (wxCommandEvent& WXUNUSED (event))
{
	wxMessageBox (aboutinfo, "About the Program",
		wxOK | wxICON_INFORMATION, this);
	return;
} /* KrakFrame::OnAbout */

void KrakFrame::OnHelpGeneral (wxCommandEvent& WXUNUSED (event))
{
	wxMessageBox(proginfo,
		"General Help", wxOK | wxICON_INFORMATION, this);
	return;
} /* KrakFrame::OnHelpGeneral */

void KrakFrame::OnHelpSpecific (wxCommandEvent& WXUNUSED (event))
{
	wxMessageBox(helpinfo,
		"Specific Help", wxOK | wxICON_INFORMATION, this);
	return;
} /* KrakFrame::OnHelpSpecific */

void KrakFrame::OnSize (wxSizeEvent& event)
{
	// call the original handler for this event
	wxFrame::OnSize (event);

#ifdef __WXGTK__
	// in GTK the canvas has to be resized explicitly
	int w, h;
	GetClientSize (&w, &h);
	if (krakCanvas)
		krakCanvas -> SetSize (w, h);
#endif
	return;
} /* KrakFrame::OnSize */

void KrakFrame::OnClose (wxCloseEvent& WXUNUSED (event))
{
	Destroy ();
	return;
} /* KrakFrame::OnClose */


// --------------------------------------------------
// --------------- matrix operations ----------------
// --------------------------------------------------

inline void makeIdentity (float m [16])
{
	m [ 0] = 1.0f;
	m [ 1] = 0.0f;
	m [ 2] = 0.0f;
	m [ 3] = 0.0f;

	m [ 4] = 0.0f;
	m [ 5] = 1.0f;
	m [ 6] = 0.0f;
	m [ 7] = 0.0f;

	m [ 8] = 0.0f;
	m [ 9] = 0.0f;
	m [10] = 1.0f;
	m [11] = 0.0f;

	m [12] = 0.0f;
	m [13] = 0.0f;
	m [14] = 0.0f;
	m [15] = 1.0f;

	return;
} /* makeIdentity */

inline void multMatrix (const float a [16], const float b [16],
	float result [16])
// Multiplicate the given two 4x4 matrices.
{
	int i, j, k;
	for (i = 0; i < 4; ++ i)
		for (j = 0; j < 4; ++ j)
		{
			result [(i << 2) + j] = 0.0f;
			for (k = 0; k < 4; ++ k)
				result [(i << 2) + j] +=
					a [(i << 2) + k] * b [(k << 2) + j];
		}
	return;
} /* multMatrix */

inline void copyMatrix (float dst [16], float src [16])
{
	dst [ 0] = src [ 0];
	dst [ 1] = src [ 1];
	dst [ 2] = src [ 2];
	dst [ 3] = src [ 3];
	dst [ 4] = src [ 4];
	dst [ 5] = src [ 5];
	dst [ 6] = src [ 6];
	dst [ 7] = src [ 7];
	dst [ 8] = src [ 8];
	dst [ 9] = src [ 9];
	dst [10] = src [10];
	dst [11] = src [11];
	dst [12] = src [12];
	dst [13] = src [13];
	dst [14] = src [14];
	dst [15] = src [15];
	return;
} /* copyMatrix */


// --------------------------------------------------
// ---------------- wxGL Main Canvas ----------------
// --------------------------------------------------

inline void KrakCanvas::Rotate (int x, int y)
{
	float step = exp (-magnification - 4);

	float newMatrix [16];
	if (x)
	{
		double cos_alpha = cos (x * step);
		double sin_alpha = sin (x * step);
		float rot_alpha [16];
		makeIdentity (rot_alpha);
		rot_alpha [ 0] = cos_alpha;
		rot_alpha [ 2] = -sin_alpha;
		rot_alpha [ 8] = sin_alpha;
		rot_alpha [10] = cos_alpha;
		multMatrix (rotation, rot_alpha, newMatrix);
	}
	
	if (y)
	{
		double cos_beta = cos (y * step);
		double sin_beta = sin (y * step);
		float rot_beta [16];
		makeIdentity (rot_beta);
		rot_beta [ 5] = cos_beta;
		rot_beta [ 6] = sin_beta;
		rot_beta [ 9] = -sin_beta;
		rot_beta [10] = cos_beta;
		if (x)
			multMatrix (newMatrix, rot_beta, rotation);
		else
			multMatrix (rotation, rot_beta, newMatrix);
	}

	if ((x && !y) || (!x && y))
		copyMatrix (rotation, newMatrix);
	
	return;
} /* KrakCanvas::Rotate */

inline void KrakCanvas::Rescale (int diff, int x, int y)
{
	if (diff)
		magnification += 0.01 * diff;

	if (x || y)
	{
		float step = exp (-magnification - 6);
		if (x)
			translation [0] += x * step;
		if (y)
			translation [1] -= y * step;
	}
	
	int w, h;
	GetClientSize (&w, &h);
	SetCurrent ();
	glViewport (0, 0, (GLint) w, (GLint) h);

	resizePicture (w, h);

	// show the new magnification exponent if necessary
	if (krakFrame -> showScaling)
		showScalingStatus ();

	return;
} /* KrakCanvas::Rescale */

KrakCanvas::KrakCanvas (wxWindow *parent, wxWindowID id,
	const wxPoint &pos, const wxSize &size, long style,
	const wxString &name, int *gl_attrib):
	wxGLCanvas (parent, id, pos, size, style, name, gl_attrib)
{
	// show the parent frame on the screen
	parent -> Show (TRUE);

	// initialize the picture
	SetCurrent ();
	glViewport (0, 0, (GLint) (size. x), (GLint) (size. y));
	initPicture (wxTheApp -> argc, wxTheApp -> argv);

	// initialize the rotation matrix
	makeIdentity (rotation);
	Rotate (37, 22);

	// set the right size of the scene projections etc.
	resizePicture (size. x, size. y);

	return;
} /* KrakCanvas::KrakCanvas */

KrakCanvas::~KrakCanvas ()
{
	return;
} /* KrakCanvas::~KrakCanvas */

void KrakCanvas::OnChar (wxKeyEvent& event)
{
	// read the code of the key pressed
	int ch = event. GetKeyCode ();

	// show the information on the key if requested to
	if (krakFrame -> showKeys)
	{
		wxString msg;
		msg. Printf ("Key: %d (%c).", ch, (char) ch);
		krakFrame -> SetStatusText (msg);
	}

	// interprete the key
	switch (ch)
	{
		case WXK_ESCAPE:
			wxTheApp -> ExitMainLoop ();
			break;
		case WXK_LEFT:
			Rotate (-15, 0);
			Refresh ();
			break;
		case WXK_RIGHT:
			Rotate (15, 0);
			Refresh ();
			break;
		case WXK_UP:
			Rotate (0, -15);
			Refresh ();
			break;
		case WXK_DOWN:
			Rotate (0, 15);
			Refresh ();
			break;
		case '+':
			Rescale (10, 0, 0);
			Refresh ();
			break;
		case '-':
			Rescale (-10, 0, 0);
			Refresh ();
			break;
		case 's':
		case 'S':
			Rescale (0, 0, 15);
			Refresh ();
			break;
		case 'w':
		case 'W':
			Rescale (0, 0, -15);
			Refresh ();
			break;
		case 'a':
		case 'A':
			Rescale (0, -15, 0);
			Refresh ();
			break;
		case 'd':
		case 'D':
			Rescale (0, 15, 0);
			Refresh ();
			break;
		case '0':
		case WXK_HOME:
			translation [0] = 0.0f;
			translation [1] = 0.0f;
			translation [2] = 0.0f;
			magnification = 0.0;
			Rescale (0, 0, 0);
			makeIdentity (rotation);
			Rotate (37, 22);
			Refresh ();
			break;
		default:
			event. Skip ();
			break;
	}
} /* KrakCanvas::OnChar */

void KrakCanvas::OnPaint (wxPaintEvent& WXUNUSED (event))
{
	wxPaintDC dc (this);

	// paint the picture
	SetCurrent ();
	drawPicture ();
	glFinish ();
	SwapBuffers ();

	// count the refresh events
	static int counter = 0;
	counter ++;

	// show status info if requested to
	if (krakFrame -> showRefresh)
	{
		wxString msg;
		msg. Printf ("Repaint window: %d", counter);
		krakFrame -> SetStatusText (msg);
	}
	
	return;
} /* KrakCanvas::OnPaint */

void KrakCanvas::OnSize (wxSizeEvent& event)
{
	// this is also necessary to update the context on some platforms
	wxGLCanvas::OnSize (event);

	// set GL viewport (not called by wxGLCanvas::OnSize on all platforms)
	// also recalculate the perspective and refresh the painting
	Rescale (0, 0, 0);
	Refresh ();
	
	return;
} /* KrakCanvas::OnSize */

void KrakCanvas::OnMouseEvents (wxMouseEvent& event)
{
	// previous mouse position
	static int mouse_x = 0, mouse_y = 0;
	
	// button states (left, right, middle)
	static bool rotating = false;
	static bool moving = false;
	static bool scaling = false;

	// show mouse information in the status bar if requested to
	if (krakFrame -> showMouse)
	{
		wxString msg;
		msg. Printf ("Mouse: %d, %d. Buttons: left is %s, "
			"right is %s, middle is %s.",
			event. GetX (), event. GetY (),
			event. LeftIsDown () ? "down" : "up",
			event. RightIsDown () ? "down" : "up",
			event. MiddleIsDown () ? "down" : "up");
		krakFrame -> SetStatusText (msg);
	}
	
	// should the screen be repainted after the mouse operations?
	bool refresh = false;
	
	// interprete the left mouse button event
	if (event. LeftIsDown ())
	{
		if (!rotating)
			rotating = true;
		else if ((event. GetX () != mouse_x) ||
			(event. GetY () != mouse_y))
		{
			Rotate (event. GetX () - mouse_x,
				event. GetY () - mouse_y);
			refresh = true;
		}
	}
	else
		rotating = false;

	// interprete the right mouse button event
	if (event. RightIsDown ())
	{
		if (!moving)
			moving = true;
		else if ((event. GetX () != mouse_x) ||
			(event. GetY () != mouse_y))
		{
		//	Rescale (mouse_y - event. GetY (), 0, 0);
			Rescale (0, event. GetX () - mouse_x,
				event. GetY () - mouse_y);
			refresh = true;
		}
	}
	else
		moving = false;

	// interprete the middle mouse button event
	if (event. MiddleIsDown ())
	{
		if (!scaling)
			scaling = true;
		else if (event. GetY () != mouse_y)
		{
			Rescale (mouse_y - event. GetY (), 0, 0);
			refresh = true;
		}
	}
	else
		scaling = false;

	// remember this mouse position for future use
	mouse_x = event. GetX ();
	mouse_y = event. GetY ();

	// refresh the picture if anything changed
	if (refresh)
		Refresh ();
		
#ifdef __WXGTK__
	// restore the focus for the panel, because the panel seems
	// to lose it and cannot capture the keyboard events
	krakPanel -> SetFocus ();
#endif

	return;
} /* KrakCanvas::OnMouseMove */

void KrakCanvas::OnEraseBackground (wxEraseEvent &WXUNUSED(event))
{
	// do nothing, to avoid flashing
	return;
} /* KrakCanvas::OnEraseBackground */


// --------------------------------------------------
// ---------------- wxGL Application ----------------
// --------------------------------------------------

bool KrakApp::OnInit ()
{
#ifdef __WXGTK__
	// set the correct locale for reading the real numbers
	setlocale (LC_NUMERIC, "C");
#endif

	krakFrame = new KrakFrame ((wxFrame *) NULL,
		"OpenGL test in wxWindows",
		wxPoint (50, 50), wxSize (500, 540));
	SetTopWindow (krakFrame);
	krakFrame -> Show (true);

#ifdef __WXGTK__
	// create a panel that will receive keyboard events
	krakPanel = new KrakPanel (krakFrame);
	krakPanel -> SetFocus ();
#endif

	// initialize the OpenGL picture within the frame
	int w, h;
	krakFrame -> GetClientSize (&w, &h);
	krakCanvas = new KrakCanvas (krakFrame, wxID_ANY, wxPoint (0, 0),
		wxSize (w, h), 0, "Krak Canvas");

	return TRUE; // FALSE iff exit the program now
} /* KrakApp::OnInit */

int KrakApp::OnExit ()
{
	deletePicture ();
	return 0;
} /* KrakApp::OnExit */

